from modelon.impact.client.client import Client
from modelon.impact.client.entities.project import (
    ContentType,
    ProjectType,
    StorageLocation,
)
from modelon.impact.client.entities.workspace import (
    AccessSettings,
    PublishedWorkspaceType,
    WorkspaceDefinition,
)
from modelon.impact.client.experiment_definition.expansion import (
    FullFactorial,
    LatinHypercube,
    Sobol,
)
from modelon.impact.client.experiment_definition.extension import (
    SimpleExperimentExtension,
)
from modelon.impact.client.experiment_definition.fmu_based import (
    SimpleFMUExperimentDefinition,
)
from modelon.impact.client.experiment_definition.model_based import (
    SimpleModelicaExperimentDefinition,
)
from modelon.impact.client.experiment_definition.modifiers import Enumeration
from modelon.impact.client.experiment_definition.operators import (
    Beta,
    Choices,
    Normal,
    Range,
    Uniform,
)
from modelon.impact.client.published_workspace_client import (
    PublishedWorkspaceAccessKind,
)

SimpleExperimentDefinition = SimpleFMUExperimentDefinition  # For compatibility



"""External result service module."""
import json
from typing import Any, Dict, Optional

from modelon.impact.client.sal.http import HTTPClient
from modelon.impact.client.sal.uri import URI


class ExternalResultService:
    def __init__(self, uri: URI, http_client: HTTPClient):
        self._base_uri = uri
        self._http_client = http_client

    def result_upload(
        self,
        workspace_id: str,
        path_to_result: str,
        label: Optional[str] = None,
        description: Optional[str] = None,
    ) -> Dict[str, Any]:
        url = (self._base_uri / "api/uploads/results").resolve()
        options: Dict[str, Any] = {
            "context": {"workspaceId": workspace_id},
        }
        if label:
            options["name"] = label
        if description:
            options["description"] = description
        with open(path_to_result, "rb") as f:
            multipart_form_data = {
                "file": f,
                "options": json.dumps(options),
            }
            return self._http_client.post_json(url, files=multipart_form_data)

    def get_uploaded_result(self, result_id: str) -> Dict[str, Any]:
        url = (self._base_uri / f"api/external-result/{result_id}").resolve()
        return self._http_client.get_json(url)

    def delete_uploaded_result(self, result_id: str) -> None:
        url = (self._base_uri / f"api/external-result/{result_id}").resolve()
        return self._http_client.delete_json(url)



"""This module contains operators for parametrizing batch runs."""
from __future__ import annotations

from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Optional, Union

from modelon.impact.client.experiment_definition.modifiers import (
    DataType,
    Enumeration,
    Modifier,
    Scalar,
    data_type_from_value,
)


class Operator(Modifier):
    """Base class for an Operator."""

    @abstractmethod
    def __str__(self) -> str:
        "Returns a string representation of the operator"

    def to_value(self) -> str:
        return str(self)


@dataclass
class Range(Operator):
    """Range operator class for parametrizing batch runs. Range(a, b, c) represents c
    linearly spaced values in the (real) interval [a, b]; b >= a. The functionality is
    analogous to numpy.linspace.

    Args:
        start_value: The start value for the sweep parameter.
        end_value: The end value for the sweep parameter.
        no_of_steps: The number of steps to intermediate steps
        to take between start_value and end_value.

    Example::

        from modelon.impact.client import Range

        fmu = model.compile().wait()
        experiment_definition = fmu.new_experiment_definition(
            custom_function).with_modifiers({'inertia1.J': 2,
            'inertia2.J': Range(0.1, 0.5, 3)})

    """

    start_value: float
    end_value: float
    no_of_steps: int

    def __str__(self) -> str:
        return f"range({self.start_value},{self.end_value},{self.no_of_steps})"

    def to_dict(self, name: str) -> dict[str, Any]:
        return {
            "kind": "range",
            "name": name,
            "start": self.start_value,
            "end": self.end_value,
            "steps": self.no_of_steps,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "Range":
        return Range(
            start_value=data["start"], end_value=data["end"], no_of_steps=data["steps"]
        )


class Choices(Operator):
    """Choices operator class for parametrizing batch runs. Choices defines a list of
    specified values a Modifier expression can take.

    Args:
        values: Variable number of numerical arguments to sweep.

    Example::

        from modelon.impact.client import Choices

        fmu = model.compile().wait()
        experiment_definition = fmu.new_experiment_definition(
            custom_function).with_modifiers({'inertia1.J': 2,
            'inertia2.J': Choices(0.1, 0.5)})

    """

    def __init__(self, *values: Scalar, data_type: Optional[DataType] = None):
        self.values = values
        self.data_type = self._resolve_data_type(values, data_type)

    @classmethod
    def _resolve_data_type(
        cls, values: tuple[Scalar, ...], given_data_type: Optional[DataType]
    ) -> DataType:
        used_data_type = given_data_type
        for value in values:
            value_type = data_type_from_value(value)

            if not used_data_type:
                used_data_type = value_type

            if cls._widen_from_int_to_real(given_data_type, used_data_type, value_type):
                used_data_type = DataType.REAL

            if used_data_type == DataType.REAL and value_type == DataType.INTEGER:
                pass  # Integer can be assigned to real, do nothing
            elif used_data_type != value_type:
                if given_data_type:
                    raise ValueError(
                        f"Choices value '{value}', not compatible "
                        f"with specified data type '{given_data_type}'"
                    )
                else:
                    raise ValueError(
                        f"Choices values are resolving to '{used_data_type}', "
                        f"which is not compatible with the value '{value}'"
                    )

        return used_data_type or DataType.REAL

    @classmethod
    def _widen_from_int_to_real(
        cls,
        given_data_type: Optional[DataType],
        used_data_type: DataType,
        value_type: DataType,
    ) -> bool:
        # If no given data type and current used is integer but there
        # is a value with type real then we say we can widen the type
        # to use real.
        return (
            given_data_type is None
            and used_data_type == DataType.INTEGER
            and value_type == DataType.REAL
        )

    def __str__(self) -> str:
        return f"choices({', '.join(map(str, self.values))})"

    def to_dict(self, name: str) -> dict[str, Any]:
        return {
            "kind": "choices",
            "name": name,
            "values": list(self.values),
            "dataType": self.data_type.value,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Choices:
        return Choices(values=data["values"], data_type=DataType(data["dataType"]))


@dataclass
class Uniform(Operator):
    """Uniform distribution class for parametrizing batch runs.For mathematical
    background, see e.g.,
    https://en.wikipedia.org/wiki/Continuous_uniform_distribution.

    Args:
        start (float): Starting value of the interval.
        end (float): End value of the interval. Requires start <= end.

    Example::

        from modelon.impact.client import Uniform

        fmu = model.compile().wait()
        experiment_definition = fmu.new_experiment_definition(
            custom_function).with_modifiers({'inertia1.J': 2,
            'inertia2.J': Uniform(0.1, 0.5)})

    """

    start: float
    end: float

    def __str__(self) -> str:
        return f"uniform({self.start},{self.end})"

    def to_dict(self, name: str) -> dict[str, Any]:
        return {
            "kind": "uniform",
            "name": name,
            "start": self.start,
            "end": self.end,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Uniform:
        return Uniform(start=data["start"], end=data["end"])


@dataclass
class Beta(Operator):
    """Beta distribution class for parametrizing batch runs. For mathematical
    background, see e.g., https://en.wikipedia.org/wiki/Beta_distribution.

    Args:
        alpha (float): 'alpha' resp. 'a' parameter of beta distribution, requires
            alpha > 0
        beta (float): 'beta' resp. 'b' parameter of beta distribution, requires
            beta > 0

    Example::

        from modelon.impact.client import Beta

        fmu = model.compile().wait()
        experiment_definition = fmu.new_experiment_definition(
            custom_function).with_modifiers({'inertia1.J': 2,
            'inertia2.J': Beta(0.1, 0.5, 3)})

    """

    alpha: float
    beta: float

    def __str__(self) -> str:
        return f"beta({self.alpha},{self.beta})"

    def to_dict(self, name: str) -> dict[str, Any]:
        return {
            "kind": "beta",
            "name": name,
            "alpha": self.alpha,
            "beta": self.beta,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Beta:
        return Beta(alpha=data["alpha"], beta=data["beta"])


@dataclass
class Normal(Operator):
    """Normal distribution class for parametrizing batch runs.For mathematical
    background, see e.g., https://en.wikipedia.org/wiki/Normal_distribution
    https://en.wikipedia.org/wiki/Truncated_normal_distribution Supports both
    the standard and truncated Normal distribution. The standard Normal
    distribution is the default, add additional start & end parameters for
    truncation.

    Args:
        mean (float): Mean resp. location of the distribution.
        variance (float): Variance of the normal distribution. Requires var > 0.
        start (float): Lower bound. Default: -inf
        end (float): Upper bound, requires end > start. Default: inf

    Example::

        from modelon.impact.client import Normal

        fmu = model.compile().wait()
        experiment_definition = fmu.new_experiment_definition(
            custom_function).with_modifiers({'inertia1.J': 2,
            'inertia2.J': Normal(0.1, 0.5)})

    """

    mean: float
    variance: float
    start: Optional[float] = None
    end: Optional[float] = None

    def __str__(self) -> str:
        start = self.start or float("-inf")
        end = self.end or float("inf")
        return f"normal({self.mean},{self.variance},{start},{end})"

    def to_dict(self, name: str) -> dict[str, Any]:
        return {
            "kind": "normal",
            "name": name,
            "mean": self.mean,
            "variable": self.variance,
            "start": self.start,
            "end": self.end,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Normal:
        return Normal(
            mean=data["mean"],
            variance=data["variance"],
            start=data["start"],
            end=data["end"],
        )


def get_operator_from_dict(
    data: dict[str, Any]
) -> Union[Range, Choices, Uniform, Beta, Normal, Enumeration, int, float, bool, str]:
    kind = data["kind"]
    if kind == "range":
        return Range.from_dict(data)
    elif kind == "choices":
        return Choices.from_dict(data)
    elif kind == "uniform":
        return Uniform.from_dict(data)
    elif kind == "beta":
        return Beta.from_dict(data)
    elif kind == "normal":
        return Normal.from_dict(data)
    elif kind == "value":
        data_type = DataType(data["dataType"])
        if data_type == DataType.ENUMERATION:
            return Enumeration(data["value"])
        else:
            return data["value"]
    raise ValueError(f"Unsupported operator kind: {kind}!")



from __future__ import annotations

import enum
import json
import logging
import os
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from modelon.impact.client.configuration import Experimental
from modelon.impact.client.entities.custom_function import CustomFunction
from modelon.impact.client.entities.experiment import Experiment
from modelon.impact.client.entities.external_result import ExternalResult
from modelon.impact.client.entities.file_uri import ModelicaResourceURI
from modelon.impact.client.entities.interfaces.workspace import WorkspaceInterface
from modelon.impact.client.entities.model import Model
from modelon.impact.client.entities.model_executable import ModelExecutable
from modelon.impact.client.entities.project import Project, VcsUri
from modelon.impact.client.exceptions import (
    NoAssociatedPublishedWorkspaceError,
    RemotePublishedWorkspaceLinkError,
)
from modelon.impact.client.experiment_definition.interfaces.definition import (
    BaseExperimentDefinition,
)
from modelon.impact.client.operations.experiment import ExperimentOperation
from modelon.impact.client.operations.external_result_import import (
    ExternalResultImportOperation,
)
from modelon.impact.client.operations.project_import import ProjectImportOperation
from modelon.impact.client.operations.workspace.conversion import (
    WorkspaceConversionOperation,
)
from modelon.impact.client.operations.workspace.exports import (
    Export,
    WorkspaceExportOperation,
)
from modelon.impact.client.operations.workspace.imports import WorkspaceImportOperation
from modelon.impact.client.sal.exceptions import HTTPError
from modelon.impact.client.sal.service import Service

if TYPE_CHECKING:
    from modelon.impact.client.operations.base import BaseOperation

logger = logging.getLogger(__name__)

ExperimentDefinition = Union[
    BaseExperimentDefinition,
    Dict[str, Any],
]


@dataclass
class AccessSettings:
    share_with_own_tenant: bool = True


@dataclass
class Reference:
    id: str


@dataclass
class ReleasedProjectReference:
    id: str
    name: str
    version: Optional[str] = None
    build: Optional[str] = None


@dataclass
class VcsReference:
    id: str
    vcs_uri: str


def _get_project_entry_reference(
    reference: Any,
) -> Union[ReleasedProjectReference, VcsReference, Reference]:
    if "name" in reference:
        return ReleasedProjectReference(
            id=reference.get("id"),
            name=reference.get("name"),
            version=reference.get("version"),
            build=reference.get("build"),
        )
    elif "vcsUri" in reference:
        return VcsReference(id=reference.get("id"), vcs_uri=reference.get("vcsUri"))
    else:
        return Reference(id=reference.get("id"))


class ProjectEntry:
    def __init__(self, data: Any) -> None:
        self._data = data

    @property
    def reference(self) -> Union[ReleasedProjectReference, VcsReference, Reference]:
        return _get_project_entry_reference(self._data.get("reference"))

    @property
    def id(self) -> str:
        return self.reference.id

    @property
    def disabled(self) -> bool:
        return self._data.get("disabled")

    @property
    def disabled_content(self) -> bool:
        return self._data.get("disabledContent")


@enum.unique
class PublishedWorkspaceType(enum.Enum):
    APP_MODE = "APP_MODE"
    ARCHIVE = "ARCHIVE"


@enum.unique
class PublishedWorkspaceUploadStatus(enum.Enum):
    INITIALIZING = "initializing"
    CREATED = "created"
    DELETING = "deleting"


@dataclass
class AppMode:
    model: str


@dataclass
class PublishedWorkspaceDefinition:
    name: str
    tenant_id: str
    size: int
    status: PublishedWorkspaceUploadStatus
    owner_username: str
    created_at: int
    app_mode: Optional[AppMode] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> PublishedWorkspaceDefinition:
        return cls(
            name=data["workspaceName"],
            tenant_id=data["tenantId"],
            size=data["size"],
            status=PublishedWorkspaceUploadStatus(data["status"]),
            owner_username=data["ownerUsername"],
            created_at=data["createdAt"],
            app_mode=AppMode(data["appMode"]["model"]) if data.get("appMode") else None,
        )


@dataclass
class PublishedWorkspaceRequester:
    id: str
    username: str

    @classmethod
    def from_dict(cls, userdata: Dict[str, Any]) -> PublishedWorkspaceRequester:
        return cls(id=userdata["id"], username=userdata["username"])


@dataclass
class PublishedWorkspaceACL:
    role_names: List[str]
    group_names: List[str]
    shared_with: List[PublishedWorkspaceRequester]
    requested_by: List[PublishedWorkspaceRequester]

    @classmethod
    def from_dict(cls, acl: Dict[str, Any]) -> PublishedWorkspaceACL:
        return cls(
            role_names=acl["roleNames"],
            group_names=acl["groupNames"],
            shared_with=[
                PublishedWorkspaceRequester.from_dict(requester)
                for requester in acl["sharedWith"]
            ],
            requested_by=[
                PublishedWorkspaceRequester.from_dict(requester)
                for requester in acl["requestedBy"]
            ],
        )


class PublishedWorkspace:
    """Class containing published workspace functionalities."""

    def __init__(
        self,
        id: str,
        definition: PublishedWorkspaceDefinition,
        service: Service,
    ):
        self._id = id
        self._definition = definition
        self._sal = service

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, PublishedWorkspace) and obj._id == self._id

    def __repr__(self) -> str:
        return f"Published Workspace with id '{self._id}'"

    @property
    def definition(self) -> PublishedWorkspaceDefinition:
        """Published workspace definition."""
        return self._definition

    @property
    def name(self) -> str:
        """Get or set the name for the published workspace.

        Example::

            # Get the published workspace name
            name = published_workspace.name

            # Set the published workspace name
            published_workspace.name = 'My workspace'

        """
        return self._definition.name

    @name.setter
    def name(self, workspace_name: str) -> None:
        self._sal.workspace.rename_published_workspace(self._id, workspace_name)
        data = self._sal.workspace.get_published_workspace(self._id)
        self._definition = PublishedWorkspaceDefinition.from_dict(data)

    @property
    def id(self) -> str:
        """Published Workspace id."""
        return self._id

    @property
    def created_at(self) -> int:
        """Published Workspace timestamp."""
        return self._definition.created_at

    def delete(self) -> None:
        """Deletes the published workspace.

        Example::

            published_workspace.delete()

        """
        self._sal.workspace.delete_published_workspace(self._id)

    @Experimental
    def grant_user_access(self, username: str) -> None:
        """Grant access for the published workspace to the requester.

        Args:
            username: Username of the user to grant access rights to.

        Example::

            published_workspace.grant_user_access('sasuke')

        """
        self._sal.workspace.grant_user_access(self._id, username)

    @Experimental
    def revoke_user_access(self, username: str) -> None:
        """Revoke access for the published workspace.

        Args:
            username: Username of the user to revoke access rights from.

        Example::

            published_workspace.revoke_user_access('naruto')

        """
        self._sal.workspace.revoke_user_access(self._id, username)

    @Experimental
    def grant_group_access(self, group_name: Optional[str] = None) -> None:
        """Grant group access for the published workspace.

        Args:
            group_name: Name of the group to grant access rights to. If not
            specified, the workspace is shared with the group the published workspace
            owner belongs to.

        Example::

            published_workspace.grant_group_access('impact-tenant-org1')

        """
        self._sal.workspace.grant_group_access(self._id, group_name)

    @Experimental
    def revoke_group_access(self, group_name: str) -> None:
        """Revoke group access for the published workspace.

        Args:
            group_name: Name of the group to revoke access rights from.

        Example::

            published_workspace.revoke_group_access('impact-tenant-org1')

        """
        self._sal.workspace.revoke_group_access(self._id, group_name)

    @Experimental
    def grant_community_access(self) -> None:
        """Grant community access for the published workspace.

        Example::

            published_workspace.grant_community_access()

        """
        self._sal.workspace.grant_community_access(self._id)

    @Experimental
    def revoke_community_access(self) -> None:
        """Revoke community access for the published workspace.

        Example::

            published_workspace.revoke_community_access()

        """
        self._sal.workspace.revoke_community_access(self._id)

    def _get_latest_local_workspace(self) -> Optional[Workspace]:
        resp = self._sal.workspace.workspaces_get(sharing_id=self._id)
        workspaces = [
            Workspace(item["id"], self._sal) for item in resp["data"]["items"]
        ]
        if not workspaces:
            return None
        if len(workspaces) > 1:
            logger.warning(
                "Multiple local copies of workspace with the same "
                "sharing ID exist. Picking the latest."
            )
            return sorted(
                workspaces,
                key=lambda x: x.definition.received_from.created_at,  # type: ignore
            )[-1]
        return workspaces[0]

    def import_to_userspace(self, update_if_available: bool = False) -> Workspace:
        """Imports a published workspace to userspace.

        Args:
            update_if_available: If true, the workspace is updated with the
            latest copy from the cloud. Default is False.

        Returns:
            The workspace class object.

        Example::
            pw_client = client.get_published_workspaces_client()
            published_workspace = pw_client.find(owner="sasuke", name="A workspace")[0]
            workspace = published_workspace.import_to_userspace()

        """
        local_workspace = self._get_latest_local_workspace()
        if local_workspace:
            logger.info(
                f"Local imports of workspace with sharing ID {self._id} exists."
            )
            local_ws_recieved_from = local_workspace.definition.received_from
            local_ws_created_at = local_ws_recieved_from.created_at  # type: ignore
            if self.created_at != local_ws_created_at:
                if not update_if_available:
                    logger.warning(
                        "A new update is available for the workspace with ID "
                        f'{local_workspace.id}. Set "update_if_available" to True '
                        "if you wish to update the workspace!"
                    )
                    return local_workspace
                logger.info("Updating the workspace to the latest published workspace.")
                updated_workspace = self._import_to_userspace(
                    overwrite_workspace_id=local_workspace.id
                )
                return updated_workspace
            else:
                logger.info(
                    f"Returning the local workspace with ID {local_workspace.id}."
                )
                return local_workspace

        logger.info(
            f"No local imports of workspace with sharing ID {self._id} exist."
            "Importing the published workspace to userspace."
        )
        return self._import_to_userspace()

    def _import_to_userspace(
        self, overwrite_workspace_id: Optional[str] = None
    ) -> Workspace:
        resp = self._sal.workspace.import_from_cloud(self._id, overwrite_workspace_id)
        return WorkspaceImportOperation[Workspace](
            resp["data"]["location"], self._sal, Workspace.from_import_operation
        ).wait()

    @Experimental
    def get_access_control_list(self) -> PublishedWorkspaceACL:
        """Returns the access control list for the published workspace.

        Returns:
            The PublishedWorkspaceACL class object.

        Example::

            pw_client = client.get_published_workspaces_client()
            published_workspace = pw_client.find(owner="sasuke", name="A workspace")[0]
            published_workspace_acl = published_workspace.get_access_control_list()

            published_workspace_shared_with = published_workspace_acl.shared_with
            published_workspace_requested_by = published_workspace_acl.requested_by

        """
        data = self._sal.workspace.get_published_workspace_acl(self.id)
        return PublishedWorkspaceACL.from_dict(data)


class OwnerData:
    def __init__(self, data: Any) -> None:
        self._data = data

    @property
    def username(self) -> str:
        return self._data["username"]

    @property
    def tenant_id(self) -> str:
        return self._data["tenantId"]


class ReceivedFrom:
    def __init__(self, data: Any) -> None:
        self._data = data

    @property
    def sharing_id(self) -> str:
        return self._data["sharingId"]

    @property
    def workspace_name(self) -> str:
        return self._data["workspaceName"]

    @property
    def owner(self) -> OwnerData:
        return OwnerData(self._data["owner"])

    @property
    def created_at(self) -> int:
        return self._data["createdAt"]


class ReceivedFromWorkspace:
    def __init__(self, workspace: Workspace) -> None:
        self._workspace = workspace
        self._received_from = self._workspace.definition.received_from
        if not self._received_from:
            raise NoAssociatedPublishedWorkspaceError(
                "The workspace has no link to a published workspace."
            )

    @property
    def metadata(self) -> ReceivedFrom:
        """Reference metadata for the published workspace."""
        assert self._received_from
        return self._received_from

    def get_workspace(self) -> Optional[PublishedWorkspace]:
        """Return the published workspace from which the workspace was imported from.
        Returns None if workspace is not linked to any published workspace.

        Returns:
            An PublishedWorkspace class object.

        Example::

            published_ws = workspace.received_from.get_workspace()

        """
        try:
            sharing_id = self.metadata.sharing_id
            data = self._workspace._sal.workspace.get_published_workspace(sharing_id)
        except HTTPError as e:
            logger.warning(
                f"Published workspace with ID: {sharing_id} doesn't exist. Cause{e}"
            )
            return None
        definition = PublishedWorkspaceDefinition.from_dict(data)
        return PublishedWorkspace(
            data["id"], definition=definition, service=self._workspace._sal
        )

    def has_updates(self) -> Optional[bool]:
        """Return True if there are updates available for the published workspace
        corresponding to the local workspace(if any) else False.

        Example::

            has_updates = workspace.received_from.has_updates()

        """
        pw = self.get_workspace()
        if not pw:
            return None
        local_ws_created_at = self.metadata.created_at  # type: ignore
        return pw.created_at != local_ws_created_at

    def _import_to_userspace(self, sharing_id: str) -> Workspace:
        resp = self._workspace._sal.workspace.import_from_cloud(
            sharing_id, self._workspace.id
        )
        return WorkspaceImportOperation[Workspace](
            resp["data"]["location"],
            self._workspace._sal,
            Workspace.from_import_operation,
        ).wait()

    def update_userspace(self) -> Workspace:
        """Returns the workspaces updated from the latest published one.

        Example::

            updated_workspace = workspace.received_from.update_userspace()

        """
        logger.info("Updating the workspace to the latest published workspace.")
        pub_ws = self.get_workspace()
        if not pub_ws:
            raise NoAssociatedPublishedWorkspaceError(
                "No published workspace found that are "
                f"associated with local workspace with ID: {self._workspace.id}"
            )
        return self._import_to_userspace(sharing_id=pub_ws.id)


class WorkspaceDefinition:
    def __init__(self, data: Any) -> None:
        self._data = data

    @property
    def name(self) -> str:
        """Workspace name."""
        return self._data.get("name")

    @property
    def format(self) -> str:
        """Workspace definition format."""
        return self._data.get("format")

    @property
    def description(self) -> str:
        """Workspace description."""
        return self._data.get("description")

    @property
    def created_by(self) -> str:
        """User ID of the workspace creator."""
        return self._data.get("createdBy")

    @property
    def default_project_id(self) -> str:
        """Project ID of the default workspace project."""
        return self._data.get("defaultProjectId")

    @property
    def received_from(self) -> Optional[ReceivedFrom]:
        """Received from information for the workspace, if imported from cloud."""
        return (
            ReceivedFrom(self._data.get("receivedFrom"))
            if self._data.get("receivedFrom")
            else None
        )

    @property
    def projects(self) -> List[ProjectEntry]:
        """List of workspace projects."""
        projects = self._data.get("projects", [])
        return [ProjectEntry(project) for project in projects]

    @property
    def dependencies(self) -> List[ProjectEntry]:
        """List of workspace dependencies."""
        dependencies = self._data.get("dependencies", [])
        return [ProjectEntry(dependency) for dependency in dependencies]

    def to_file(self, path: str) -> str:
        """Writes the workspace definition as a JSON file.

        Args:
            path: The path of the folder to store the file at.

        Returns:
            The file path of the created JSON file.

        Example::

            file = workspace.to_file('/home/workspace/definition')

        """
        os.makedirs(path, exist_ok=True)
        definition_path = os.path.join(path, self.name + ".json")
        with open(definition_path, "w", encoding="utf-8") as f:
            json.dump(self._data, f, indent=4)
        return definition_path

    @classmethod
    def from_file(cls, path: str) -> WorkspaceDefinition:
        """Constructs WorkspaceDefinition class from JSON file.

        Args:
            path: The path of the JSON file.

        Returns:
            The WorkspaceDefinition class object.

        Example::

            definition = workspace.from_file('/home/definition.json')

        """
        with open(path) as json_file:
            data = json.load(json_file)
        return cls(data)

    def to_dict(self) -> Dict[str, Any]:
        """Returns the workspace definition as a dict."""
        return self._data


class Workspace(WorkspaceInterface):
    """Class containing Workspace functionalities."""

    def __init__(
        self,
        workspace_id: str,
        service: Service,
    ):
        self._workspace_id = workspace_id
        self._sal = service

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, Workspace) and obj._workspace_id == self._workspace_id

    def __repr__(self) -> str:
        return f"Workspace with id '{self._workspace_id}'"

    @property
    def id(self) -> str:
        """Workspace id."""
        return self._workspace_id

    @property
    def size(self) -> float:
        """Workspace size in bytes."""
        return self._sal.workspace.workspace_get(workspace_id=self.id, size_info=True)[
            "sizeInfo"
        ]["total"]

    @property
    def definition(self) -> WorkspaceDefinition:
        """Workspace definition."""
        definition = self._sal.workspace.workspace_get(
            workspace_id=self.id, size_info=False
        )["definition"]
        return WorkspaceDefinition(definition)

    @property
    def name(self) -> str:
        """Workspace name."""
        return self.definition.name

    def rename(self, new_name: str) -> None:
        """Renames a workspace.

        Args:
            name: The new name for the workspace.

        Example::

            workspace.rename('renamed workspace')

        """
        workspace_data = self._sal.workspace.workspace_get(
            workspace_id=self.id, size_info=False
        )
        workspace_data["definition"]["name"] = new_name
        self._sal.workspace.update_workspace(self.id, workspace_data)

    def get_custom_function(self, name: str) -> CustomFunction:
        """Returns a CustomFunction class object.

        Args:
            name: The name of the custom function.

        Returns:
            The CustomFunction class object.

        Example::

            workspace.get_custom_function('dynamic')

        """
        custom_function = self._sal.custom_function.custom_function_get(
            self._workspace_id, name
        )
        return CustomFunction(
            self._workspace_id,
            custom_function["name"],
            custom_function["parameters"],
            self._sal,
        )

    def get_custom_functions(self) -> List[CustomFunction]:
        """Returns a list of CustomFunctions class objects.

        Returns:
            A list of CustomFunction class objects.

        Example::

            workspace.get_custom_functions()

        """
        custom_functions = self._sal.custom_function.custom_functions_get(
            self._workspace_id
        )
        return [
            CustomFunction(
                self._workspace_id,
                custom_function["name"],
                custom_function["parameters"],
                self._sal,
            )
            for custom_function in custom_functions["data"]["items"]
        ]

    def delete(self) -> None:
        """Deletes a workspace.

        Example::

            workspace.delete()

        """
        self._sal.workspace.workspace_delete(self._workspace_id)

    def upload_result(
        self,
        path_to_result: str,
        label: Optional[str] = None,
        description: Optional[str] = None,
    ) -> ExternalResultImportOperation:
        """Uploads a '.mat' result file to the workspace.

        Args:
            path_to_result: The path for the result file to be imported.
            label: The label of the result file. Default: None.
            description: The description of the result file. Default: None.

        Example::

            workspace.upload_result('A.mat')
            workspace.upload_result('B.mat', label = "result_for_PID.mat",
            description = "This is a result file for PID controller")

        """
        resp = self._sal.external_result.result_upload(
            self._workspace_id, path_to_result, label=label, description=description
        )
        return ExternalResultImportOperation[ExternalResult](
            resp["data"]["location"], self._sal, ExternalResult.from_operation
        )

    def export(
        self,
        publish: bool = False,
        class_path: Optional[str] = None,
        access: Optional[AccessSettings] = None,
    ) -> WorkspaceExportOperation:
        """Exports the workspace as a binary compressed archive. Similar to
        :obj:`~modelon.impact.client.entities.workspace.Workspace.download`, but gives
        more control for getting the workspace async. Returns an
        modelon.impact.client.operations.workspace.exports .WorkspaceExportOperation
        class object. The binary archive is stored to cloud storage if the publish
        argument is set to True.

        Args:
            publish: To export the workspace and save it to cloud storage.
            class_path: The Modelica class path of the model. If specified,
                the workspace is exported in app mode.
            access: The access control settings for the workspace.

        Returns:
            An WorkspaceExportOperation class object.

        Example::

            path = workspace.export().wait().download_as('/home/workspace.zip')

            # Publish a workspace
            workspace.export(publish=True,
                class_path='Modelica.Blocks.Examples.PID_Controller').wait()

            # Publish a workspace without sharing with own tenant
            from modelon.impact.client import AccessSettings

            workspace.export(
                publish=True,
                class_path='Modelica.Blocks.Examples.PID_Controller',
                access=AccessSettings(share_with_own_tenant=False)
            ).wait()

        """
        if access:
            access_settings = {"shareWithOwnTenant": access.share_with_own_tenant}
        else:
            access_settings = None
        resp = self._sal.workspace.workspace_export_setup(
            self._workspace_id,
            publish,
            class_path,
            access_settings,
        )
        return WorkspaceExportOperation[Workspace](
            resp["data"]["location"], self._sal, Export.from_operation
        )

    def download(self, path: str) -> str:
        """Downloads the workspace as a binary compressed archive. Returns the local
        path to the downloaded workspace archive. Similar to
        :obj:`~modelon.impact.client.entities.workspace.Workspace.export`, but does the
        entire setup and download in one go.

        Args:

            path: The local path to store the downloaded workspace.

        Returns:
            Local path to the downloaded workspace archive.

        Example::

            workspace.download(path)

        """
        ws_path = os.path.join(path, self._workspace_id + ".zip")
        ops = self.export().wait()
        return ops.download_as(ws_path)

    def get_model(self, class_name: str, project: Optional[Project] = None) -> Model:
        """Returns a Model class object.

        Args:
            class_name: The Modelica class path of the model.
            project: Project class object

        Returns:
            Model class object.

        Example::

            workspace.get_model(class_name)

        """
        project = project or self.get_default_project()
        return Model(class_name, self._workspace_id, project.id, self._sal)

    def get_fmus(self) -> List[ModelExecutable]:
        """Returns a list of ModelExecutable class objects.

        Returns:
            List of ModelExecutable class objects.

        Example::

            workspace.get_fmus()

        """
        resp = self._sal.workspace.fmus_get(self._workspace_id)
        return [
            ModelExecutable(self._workspace_id, item["id"], self._sal, item)
            for item in resp["data"]["items"]
        ]

    def get_fmu(self, fmu_id: str) -> ModelExecutable:
        """Returns a ModelExecutable class object.

        Returns:
            ModelExecutable class object.

        Example::

            workspace.get_fmu(fmu_id)

        """
        resp = self._sal.workspace.fmu_get(self._workspace_id, fmu_id)
        return ModelExecutable(self._workspace_id, resp["id"], self._sal, resp)

    def get_experiments(self, class_path: Optional[str] = None) -> List[Experiment]:
        """Returns a list of Experiment class objects.

        Parameters:

            class_path --
                The modelica class path. If given, only the experiments
                generated for model with the specified class path
                are returned.

        Returns:
            List of Experiment class objects.

        Example::

            workspace.get_experiments()

        """
        resp = self._sal.workspace.experiments_get(self._workspace_id, class_path)
        return [
            Experiment(self._workspace_id, item["id"], self._sal, item)
            for item in resp["data"]["items"]
        ]

    def get_experiment(self, experiment_id: str) -> Experiment:
        """Returns an Experiment class object.

        Args:
            experiment_id:
                The ID of the experiment.

        Returns:
            Experiment class object.

        Example::

            workspace.get_experiment(experiment_id)

        """
        resp = self._sal.workspace.experiment_get(self._workspace_id, experiment_id)
        return Experiment(self._workspace_id, resp["id"], self._sal, resp)

    def create_experiment(
        self,
        definition: ExperimentDefinition,
        user_data: Optional[Dict[str, Any]] = None,
    ) -> Experiment:
        """Creates an experiment. Returns an Experiment class object.

        Args:
            definition: A parametrized experiment definition class of type
                SimpleModelicaExperimentDefinition or SimpleFMUExperimentDefinition.
            user_data:
                Optional dictionary object with custom data to attach to the experiment.

        Returns:
            Experiment class object.

        Example::

            workspace.create_experiment(definition)

        """
        if isinstance(
            definition,
            BaseExperimentDefinition,
        ):
            definition_as_dict = definition.to_dict()
        elif isinstance(definition, dict):
            definition_as_dict = definition
        else:
            raise TypeError(
                "Definition object must either be a dictionary or an instance of either"
                " SimpleModelicaExperimentDefinition or SimpleFMUExperimentDefinition"
                " class!"
            )
        resp = self._sal.workspace.experiment_create(
            self._workspace_id, definition_as_dict, user_data
        )
        return Experiment(self._workspace_id, resp["experiment_id"], self._sal)

    def execute(
        self,
        definition: ExperimentDefinition,
        user_data: Optional[Dict[str, Any]] = None,
    ) -> ExperimentOperation:
        """Executes an experiment.

        Args:
            definition:
                An experiment definition class instance of
                SimpleFMUExperimentDefinition
                or SimpleModelicaExperimentDefinition
                or a dictionary object containing the definition.
            user_data:
                Optional dictionary object with custom data to attach to the experiment.


        Returns:
            An ExperimentOperation class object.

        Example::

            experiment_ops = workspace.execute(definition)
            experiment_ops.cancel()
            experiment_ops.status
            experiment_ops.wait()

        """
        exp_id = self.create_experiment(definition, user_data).id
        return ExperimentOperation[Experiment](
            self._workspace_id,
            self._sal.experiment.experiment_execute(self._workspace_id, exp_id),
            self._sal,
            Experiment.from_operation,
        )

    def get_projects(
        self, vcs_info: bool = True, include_disabled: bool = False
    ) -> List[Project]:
        """Return the list of projects for a workspace.

        Args:
            vcs_info: If True, the versioning details are returned for the
            project(if under version control).
            include_disabled: If True, projects disabled in the workspace
            are also listed.

        Returns:
            A list of Project class objects.

        Example::

            projects = workspace.get_projects()

        """
        resp = self._sal.workspace.projects_get(
            self._workspace_id, vcs_info=vcs_info, include_disabled=include_disabled
        )
        projects = [
            Project(
                item["id"],
                item["projectType"],
                item["storageLocation"],
                VcsUri.from_dict(item["vcsUri"]) if item.get("vcsUri") else None,
                self._sal,
            )
            for item in resp["data"]["items"]
        ]
        return projects

    def get_dependencies(
        self, vcs_info: bool = True, include_disabled: bool = False
    ) -> List[Project]:
        """Return the list of project dependencies for a workspace.

        Args:
            vcs_info: If True, the versioning details are returned for the
            project(if under version control).
            include_disabled: If True, projects disabled in the workspace
            are also listed.

        Returns:
            A list of Project class object.

        Example::

            dependencies = workspace.get_dependencies()

        """
        resp = self._sal.workspace.dependencies_get(
            self._workspace_id, vcs_info, include_disabled
        )
        return [
            Project(
                item["id"],
                item["projectType"],
                item["storageLocation"],
                VcsUri.from_dict(item["vcsUri"]) if item.get("vcsUri") else None,
                self._sal,
            )
            for item in resp["data"]["items"]
        ]

    def create_project(self, name: str) -> Project:
        """Creates a new project in the workspace.

        Returns:
            A Project class object.

        Example::

            project = workspace.create_project("test")

        """
        resp = self._sal.workspace.project_create(self._workspace_id, name)
        return Project(
            resp["id"],
            resp["projectType"],
            resp["storageLocation"],
            VcsUri.from_dict(resp["vcsUri"]) if resp.get("vcsUri") else None,
            self._sal,
        )

    def get_default_project(self) -> Project:
        """Return the default project for a workspace.

        Returns:
            An Project class object.

        Example::

            project = workspace.get_default_project()

        """
        default_project_id = self.definition.default_project_id
        if not default_project_id:
            raise ValueError(
                f"No default project exists for the workspace {self._workspace_id}!"
            )
        resp = self._sal.project.project_get(
            default_project_id,
            vcs_info=True,
            size_info=False,
        )
        return Project(
            resp["id"],
            resp["projectType"],
            resp["storageLocation"],
            VcsUri.from_dict(resp["vcsUri"]) if resp.get("vcsUri") else None,
            self._sal,
        )

    def get_shared_definition(self, strict: bool = False) -> WorkspaceDefinition:
        """Return the shared definition for a workspace.

        Args:
            strict: If True, the version control URIs for workspace projects and
                dependencies are to specific commits or not.

        Returns:
            An WorkspaceDefinition class object.

        Example::

            definition = workspace.get_shared_definition()

        """
        return WorkspaceDefinition(
            self._sal.workspace.shared_definition_get(
                self._workspace_id, strict=strict
            )["definition"]
        )

    def import_project_from_zip(self, path_to_project: str) -> ProjectImportOperation:
        """Imports a Project from a compressed(.zip) project file and adds it to the
        workspace. Returns the project class object.

        Args:
            path_to_project: The path for the compressed project(.zip) to be uploaded.

        Returns:
            A ProjectImportOperation class object.

        Example::

            workspace.import_project_from_zip(path_to_project).wait()

        """
        resp = self._sal.workspace.import_project_from_zip(
            self._workspace_id, path_to_project
        )
        return ProjectImportOperation[Project](
            resp["data"]["location"], self._sal, Project.from_operation
        )

    def import_dependency_from_zip(
        self, path_to_dependency: str
    ) -> ProjectImportOperation:
        """Imports a Project dependency from a compressed(.zip) project file and adds it
        to the workspace. Returns the project class object.

        Args:
            path_to_dependency: The path for the compressed project(.zip) to be
            uploaded.

        Returns:
            An ProjectImportOperation class object.

        Example::

            workspace.import_dependency_from_zip(path_to_project).wait()

        """
        resp = self._sal.workspace.import_dependency_from_zip(
            self._workspace_id, path_to_dependency
        )
        return ProjectImportOperation[Project](
            resp["data"]["location"], self._sal, Project.from_operation
        )

    @classmethod
    def from_import_operation(
        cls, operation: BaseOperation[Workspace], **kwargs: Any
    ) -> Workspace:
        assert isinstance(operation, WorkspaceImportOperation)
        return cls(**kwargs, service=operation._sal)

    @classmethod
    def from_conversion_operation(
        cls, operation: BaseOperation[Workspace], **kwargs: Any
    ) -> Workspace:
        assert isinstance(operation, WorkspaceConversionOperation)
        return cls(**kwargs, service=operation._sal)

    @property
    def received_from(self) -> ReceivedFromWorkspace:
        """Returns an instance of ReceivedFromWorkspace class."""
        return ReceivedFromWorkspace(self)

    def get_published_workspace(
        self, model_name: Optional[str] = None
    ) -> Optional[PublishedWorkspace]:
        """Returns the published workspace with the name matching this workspace.

        Args:
            model_name: Class path of the Modelica model. If specified,
             an app mode published workspace corresponding to the local workspace
             is returned(if it exists).

        Returns:
            An PublishedWorkspace class object.

        Example::

            published_ws = workspace.get_published_workspace()

        """
        recieved_from = self.definition.received_from
        if recieved_from:
            raise RemotePublishedWorkspaceLinkError(
                "This workspace is imported from published workspace with ID:"
                f"{recieved_from.sharing_id}. Use received_from.get_workspace()"
                " to retrieve the information"
            )
        user = self._sal.users.get_me()["data"]
        published_workspaces = self._sal.workspace.get_published_workspaces(
            name=self.name,
            owner_username=user["username"],
            type=PublishedWorkspaceType.APP_MODE.value
            if model_name
            else PublishedWorkspaceType.ARCHIVE.value,
        )["data"]["items"]
        if model_name:
            published_workspaces = [
                pw
                for pw in published_workspaces
                if pw.get("appMode", {}).get("model") == model_name
            ]
        if not published_workspaces:
            logger.warning(
                "No published workspace corresponding to the local workspace with"
                f" ID: {self.id} exist!"
            )
            return None
        return PublishedWorkspace(
            published_workspaces[0]["id"],
            definition=PublishedWorkspaceDefinition.from_dict(published_workspaces[0]),
            service=self._sal,
        )

    @Experimental
    def get_modelica_resource_uri(
        self, library: str, resource_path: str
    ) -> ModelicaResourceURI:
        """Returns a ModelicaResourceURI class.

        Returns:
            The ModelicaResourceURI class object.

        Example::

            modelica_resource_ref = workspace.get_modelica_resource_uri(
                library,
                resource_path
            )

        """
        return ModelicaResourceURI(library=library, resource_path=resource_path)



import logging
from dataclasses import asdict, dataclass
from typing import Any, Dict, Optional

from modelon.impact.client.experiment_definition.interfaces.expansion import (
    ExpansionAlgorithm,
)

logger = logging.getLogger(__name__)


@dataclass
class LatinHypercube(ExpansionAlgorithm):
    """LatinHypercube expansion class. Produces <samples> cases, picks a random value
    from each modifier expression for each case. The resulting cases are orthogonal,
    i.e., the values of a given modifier expression do not repeat. The exception are
    singular modifiers, e.g., experiment_definition.with_modifiers({'PI.k': 10}), these
    result in the same value for all cases. Singular modifiers do not affect the result
    (with respect to the seed) in the resulting experiment.

    Args:
        samples (int):
            Positive integer; number of cases the experiment will produce.
        seed (int) :
            Using the same seed will result in the same output for an experiment with
            the same modifiers. If not set or None: picks a random seed. Must be a
            non-negative integer. Default: None.

    Example::

        from modelon.impact.client import LatinHypercube, Beta, Normal

        model = workspace.get_model("Modelica.Blocks.Examples.PID_Controller")
        experiment_definition = model.new_experiment_definition(
            custom_function).with_modifiers({'inertia1.J': Beta(0.1, 0.9),
            'inertia2.J': Normal(0.1, 0.5)}).with_expansion(LatinHypercube(5,1))

    """

    samples: int
    seed: Optional[int] = None

    def __str__(self) -> str:
        return "LATINHYPERCUBE"

    def get_parameters_as_dict(self) -> Optional[Dict[str, Any]]:
        return asdict(self)


class FullFactorial(ExpansionAlgorithm):
    """Full-factorial expansion class. Creates experiment with all possible
    combinations of the input modifiers expressions. Supported Operator
    expressions are: Range and Choices. The size of an experiment with
    FullFactorial expansion is the product of the modifier operator length.
    This number can grow very rapidly if using a lot of modifiers.

    Example::

        from modelon.impact.client import FullFactorial, Range

        model = workspace.get_model("Modelica.Blocks.Examples.PID_Controller")

        experiment_definition = model.new_experiment_definition(custom_function)
        .with_modifiers({'inertia1.J': Range(0.1, 0.9, 2)})
        .with_expansion(FullFactorial())

        OR

        # Full-factorial is the default and does not need to be specified explicitly.
        experiment_definition = model.new_experiment_definition(custom_function)
        .with_modifiers({'inertia1.J': Range(0.1, 0.9, 2)})

    """

    def __str__(self) -> str:
        return "FULLFACTORIAL"

    def get_parameters_as_dict(self) -> Optional[Dict[str, Any]]:
        return None


@dataclass
class Sobol(ExpansionAlgorithm):
    """Expansion method based on the Sobol sequence. The Sobol sequence is a minimal
    discrepancy quasi-random sampling methods and suitable for achieving a good coverage
    of the design space. Singular Modifiers, e.g.,
    experiment_definition.with_modifiers({'PI.k': 10}), do not affect the result in the
    resulting experiment.

    Args:
        samples (int):
            Positive integer; number of cases the Experiment will produce.
            Expansions where the number of samples is a power of 2 yield
            additional balances properties.

    Example::

        from modelon.impact.client import Sobol, Beta, Normal

        model = workspace.get_model("Modelica.Blocks.Examples.PID_Controller")
        experiment_definition = model.new_experiment_definition(
            custom_function).with_modifiers({'inertia1.J': Beta(0.1, 0.9),
            'inertia2.J': Normal(0.1, 0.5)}).with_expansion(Sobol(5))

    """

    samples: int

    def __str__(self) -> str:
        return "SOBOL"

    def get_parameters_as_dict(self) -> Optional[Dict[str, Any]]:
        return asdict(self)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict

if TYPE_CHECKING:
    from modelon.impact.client.sal.service import Service


class CaseInterface(ABC):
    def __init__(
        self,
        case_id: str,
        workspace_id: str,
        exp_id: str,
        service: Service,
        info: Dict[str, Any],
    ):
        self._case_id = case_id
        self._workspace_id = workspace_id
        self._exp_id = exp_id
        self._sal = service
        self._info = info

    @property
    @abstractmethod
    def id(self) -> str:
        """Case id."""
        raise NotImplementedError

    @property
    def experiment_id(self) -> str:
        """Experiment id."""
        raise NotImplementedError


class CaseReference(CaseInterface):
    def __init__(
        self,
        case_id: str,
        workspace_id: str,
        exp_id: str,
        service: Service,
        info: Dict[str, Any],
    ):
        super().__init__(case_id, workspace_id, exp_id, service, info)

    @property
    def id(self) -> str:
        """Case id."""
        return self._case_id

    @property
    def experiment_id(self) -> str:
        """Experiment id."""
        return self._exp_id



from enum import Enum


class CaseStatus(Enum):
    """Class representing an enumeration for the possible Case run info states."""

    SUCCESSFUL = "successful"
    """Status for case that has been executed successfully."""

    FAILED = "failed"
    """Status for case execution that has failed."""

    CANCELLED = "cancelled"
    """Status for experiment execution that has not yet started."""

    NOT_STARTED = "not_started"
    """Status for experiment execution that has not yet started."""

    STARTED = "started"
    """Status for case that has  started execution."""


class ExperimentStatus(Enum):
    """Class representing an enumeration for the possible experiment run info states."""

    NOTSTARTED = "not_started"
    """Status for experiment execution that has not yet started."""

    CANCELLED = "cancelled"
    """Status for a cancelled experiment execution."""

    DONE = "done"
    """Status for a completed experiment execution."""

    FAILED = "failed"
    """Status for a failed experiment execution."""


class ModelExecutableStatus(Enum):
    """Class representing an enumeration for the possible model-executable run info
    states."""

    NOTSTARTED = "not_started"
    """Status for compilation that has not yet started."""

    CANCELLED = "cancelled"
    """Status for a cancelled compilation."""

    SUCCESSFUL = "successful"
    """Status for a successful compilation."""

    FAILED = "failed"
    """Status for a failed compilation."""



import logging
import os
from getpass import getpass
from typing import List, Optional

logger = logging.getLogger(__name__)


class CredentialManager:
    def __init__(
        self,
        file_id: str = "api.key",
        env_names: List[str] = ["MODELON_IMPACT_CLIENT_API_KEY"],
        interactive_help_text: str = "Enter Modelon Impact API key:",
    ):
        self._file_id = file_id
        self._env_names = env_names
        self._interactive_help_text = interactive_help_text

    def get_key_from_env(self) -> Optional[str]:
        for env_name in self._env_names:
            env_value = os.environ.get(env_name)
            if env_value:
                return env_value
        return None

    def _cred_file(self) -> str:
        home_dir = os.path.expanduser("~")
        return os.path.join(home_dir, ".impact", self._file_id)

    def get_key_from_file(self) -> Optional[str]:
        credentials_file = self._cred_file()
        if not os.path.isfile(credentials_file):
            return None

        with open(credentials_file, "r") as credentials:
            return credentials.read().strip()

    def get_key_from_prompt(self) -> str:
        key = getpass(self._interactive_help_text)
        if key == "\x16":
            # This is a getpass Windows bug...
            logger.error("Does not support Ctrl+V on Windows")
            return self.get_key_from_prompt()

        return key

    def write_key_to_file(self, api_key: str) -> None:
        credentials_file = self._cred_file()
        os.makedirs(os.path.dirname(credentials_file), exist_ok=True)
        with open(credentials_file, "w") as credentials:
            credentials.write(str(api_key))

    def get_key(self, interactive: bool = False) -> Optional[str]:
        api_key = self.get_key_from_env() or self.get_key_from_file()
        if api_key or not interactive:
            return api_key

        return self.get_key_from_prompt()



from __future__ import annotations

import enum
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Union
from urllib.parse import urlparse

from typing_extensions import assert_never


@enum.unique
class URISchema(enum.Enum):
    """Supported Version Control System services."""

    MODELICA = "MODELICA"
    IMPACT_ARTIFACT = "IMPACT-ARTIFACT"

    @classmethod
    def from_str(cls, value: str) -> URISchema:
        for member in cls:
            if member.value == value.upper():
                return member
        raise ValueError(f"Incorrect schema in URI: {value!r}")


@dataclass
class URI:
    schema: URISchema
    netloc: str
    path: str

    @classmethod
    def from_str(cls, uri_str: str) -> URI:
        """Parse URI str into this class instance.

        Args:
            uri_str (str): <scheme>://<netloc>/<path>

        Returns:
            URI: This class instance.

        """
        parsed_url = urlparse(uri_str)
        netloc = parsed_url.netloc
        scheme = parsed_url.scheme
        path = parsed_url.path.lstrip("/")
        uri_schema = URISchema.from_str(scheme)
        return cls(uri_schema, netloc, path)

    def __str__(self) -> str:
        return f"{self.schema.value.lower()}://{self.netloc}/{self.path}"


class FileURI(ABC):
    def __str__(self) -> str:
        return str(self.uri)

    @property
    @abstractmethod
    def uri(self) -> URI:
        """URI class."""
        raise NotImplementedError


class ModelicaResourceURI(FileURI):
    def __init__(self, library: str, resource_path: str) -> None:
        self._library = library
        self._resource_path = resource_path

    @property
    def uri(self) -> URI:
        return URI.from_str(f"modelica://{self._library}/{self._resource_path}")

    @classmethod
    def from_uri(cls, uri: URI) -> ModelicaResourceURI:
        return cls(uri.netloc, uri.path)


class CustomArtifactURI(FileURI):
    def __init__(self, experiment_id: str, case_id: str, artifact_id: str) -> None:
        self._experiment_id = experiment_id
        self._case_id = case_id
        self._artifact_id = artifact_id

    @property
    def uri(self) -> URI:
        return URI.from_str(
            f"impact-artifact://workspace/{self._experiment_id}/{self._case_id}/"
            f"{self._artifact_id}"
        )

    @classmethod
    def from_uri(cls, uri: URI) -> CustomArtifactURI:
        experiment_id, case_id, artifact_id = uri.path.split("/")
        return cls(experiment_id, case_id, artifact_id)


def get_resource_URI_from_str(
    uri_str: str,
) -> Union[ModelicaResourceURI, CustomArtifactURI]:
    file_uri = URI.from_str(uri_str)
    if file_uri.schema == URISchema.MODELICA:
        return ModelicaResourceURI.from_uri(file_uri)
    elif file_uri.schema == URISchema.IMPACT_ARTIFACT:
        return CustomArtifactURI.from_uri(file_uri)
    else:
        assert_never(file_uri.schema)



from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Dict, Iterator, List

from modelon.impact.client.entities.asserts import assert_variable_in_result
from modelon.impact.client.operations.base import BaseOperation
from modelon.impact.client.operations.case_result import CaseResultImportOperation

if TYPE_CHECKING:
    from modelon.impact.client.sal.experiment import ExperimentService
    from modelon.impact.client.sal.service import Service


def _create_result_dict(
    variables: List[str],
    workspace_id: str,
    exp_id: str,
    case_id: str,
    exp_sal: ExperimentService,
    last_point_only: bool,
) -> Dict[str, Any]:
    response = exp_sal.trajectories_get(
        workspace_id, exp_id, variables, last_point_only
    )
    case_index = int(case_id.split("_")[1])
    data = {
        variable: response[i][case_index - 1] for i, variable in enumerate(variables)
    }
    return data


class Result(Mapping):
    """Result class containing base functionality."""

    def __init__(
        self,
        variables: List[str],
        case_id: str,
        workspace_id: str,
        exp_id: str,
        service: Service,
    ):
        self._case_id = case_id
        self._workspace_id = workspace_id
        self._exp_id = exp_id
        self._sal = service
        self._variables = variables

    def __getitem__(self, key: str) -> Any:
        assert_variable_in_result([key], self._variables)
        # TODO: Implement support for last point only
        response = self._sal.experiment.case_trajectories_get(
            self._workspace_id, self._exp_id, self._case_id, [key], False
        )
        return response[0]

    def __iter__(self) -> Iterator[Any]:
        data = _create_result_dict(
            self._variables,
            self._workspace_id,
            self._exp_id,
            self._case_id,
            self._sal.experiment,
            False,
        )
        return data.__iter__()

    def __len__(self) -> int:
        return self._variables.__len__()

    def keys(self):  # type: ignore
        return self._variables

    @classmethod
    def from_operation(cls, operation: BaseOperation[Result], **kwargs: Any) -> Result:
        assert isinstance(operation, CaseResultImportOperation)
        return cls(**kwargs, service=operation._sal)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from modelon.impact.client import exceptions
from modelon.impact.client.entities.interfaces.case import CaseInterface
from modelon.impact.client.entities.interfaces.experiment import ExperimentInterface
from modelon.impact.client.entities.interfaces.external_result import (
    ExternalResultInterface,
)
from modelon.impact.client.experiment_definition.asserts import (
    assert_valid_args,
    assert_valid_case_modifiers,
    assert_valid_extensions,
    validate_initialize_from,
)
from modelon.impact.client.experiment_definition.extension import (
    SimpleExperimentExtension,
)
from modelon.impact.client.experiment_definition.interfaces.definition import (
    BaseExperimentDefinition,
)
from modelon.impact.client.experiment_definition.modifiers import (
    ensure_as_modifier,
    modifiers_to_dict,
)
from modelon.impact.client.experiment_definition.util import (
    case_to_identifier_dict,
    custom_function_parameters_to_dict,
    get_options,
)

if TYPE_CHECKING:
    from modelon.impact.client.entities.case import Case
    from modelon.impact.client.entities.custom_function import CustomFunction
    from modelon.impact.client.entities.experiment import Experiment
    from modelon.impact.client.entities.external_result import ExternalResult
    from modelon.impact.client.entities.model_executable import ModelExecutable
    from modelon.impact.client.options import SimulationOptions, SolverOptions

    CaseOrExperimentOrExternalResult = Union[Case, Experiment, ExternalResult]
    SimulationOptionsOrDict = Union[SimulationOptions, Dict[str, Any]]
    SolverOptionsOrDict = Union[SolverOptions, Dict[str, Any]]

logger = logging.getLogger(__name__)


def _assert_successful_compilation(fmu: ModelExecutable) -> None:
    if not fmu.is_successful():
        raise exceptions.OperationFailureError(
            "Compilation process has failed! See the log for more info!"
        )


class SimpleFMUExperimentDefinition(BaseExperimentDefinition):
    """A simple experiment definition class for defining experiments.

    Args:
        fmu: The FMU to be executed for this experiment.
        custom_function: The custom function to use for this experiment.

        solver_options: The solver options to use for this experiment.
            By default, the options is set to None, which means the
            default options for the custom_function input is used.
        simulation_options: The simulation_options to use for this
            experiment. By default, the options is set to None, which
            means the default options for the custom_function input is used.
        simulation_log_level: Simulation log level for this experiment.
            Default is 'WARNING'.
        initialize_from: Optional entity to initialize from. An instance of
            Case or Experiment or ExternalResult. Default: None

    Example::

        fmu = model.compile().wait()
        simulation_options = custom_function.get_simulation_options()
        .with_values(ncp=500)
        solver_options = {'atol':1e-8}
        simulate_def = fmu.new_experiment_definition(custom_function,
        solver_options, simulation_options)
        simulate_def.to_dict()

    """

    def __init__(
        self,
        fmu: ModelExecutable,
        custom_function: CustomFunction,
        solver_options: Optional[SolverOptionsOrDict] = None,
        simulation_options: Optional[SimulationOptionsOrDict] = None,
        simulation_log_level: str = "WARNING",
        initialize_from: Optional[CaseOrExperimentOrExternalResult] = None,
    ):
        assert_valid_args(
            fmu=fmu,
            custom_function=custom_function,
            solver_options=solver_options,
            simulation_options=simulation_options,
        )
        _assert_successful_compilation(fmu)
        self._fmu = fmu
        self._custom_function = custom_function
        self._solver_options = get_options(
            custom_function.get_solver_options, solver_options
        )
        self._simulation_options = get_options(
            custom_function.get_simulation_options, simulation_options
        )

        self._simulation_log_level = simulation_log_level
        self._initialize_from = initialize_from
        self._variable_modifiers = fmu._variable_modifiers()
        self._extensions: List[SimpleExperimentExtension] = []

    @property
    def fmu(self) -> ModelExecutable:
        """Returns the ModelExecutable class."""
        return self._fmu

    @property
    def modifiers(self) -> Dict[str, Any]:
        """Returns the variable modifiers dict."""
        return self._variable_modifiers

    @property
    def simulation_options(self) -> Dict[str, Any]:
        """Returns the simulation options as a dict."""
        return self._simulation_options

    @property
    def solver_options(self) -> Dict[str, Any]:
        """Returns the solver options as a dict."""
        return self._solver_options

    @property
    def simulation_log_level(self) -> str:
        """Returns the simulation log level."""
        return self._simulation_log_level

    @property
    def custom_function(self) -> CustomFunction:
        """Returns the custom function class."""
        return self._custom_function

    @property
    def extensions(self) -> List[SimpleExperimentExtension]:
        """Returns the list of experiment extensions."""
        return self._extensions

    @property
    def initialize_from(self) -> Optional[CaseOrExperimentOrExternalResult]:
        """Returns the case, experiment or result the experiment definition was
        initialized from."""
        return self._initialize_from

    def validate(self) -> None:
        add = set(self._variable_modifiers.keys()) - set(
            self._fmu.get_settable_parameters()
        )
        if add:
            raise KeyError(
                f"Paramter(s) '{', '.join(add)}' {'are' if len(add)>1 else 'is'} "
                "not a valid parameter modifier! Call the get_settable_parameters() "
                "method on the fmu to view the list of settable parameters."
            )

    def with_modifiers(
        self, modifiers: Optional[Dict[str, Any]] = None, **modifiers_kwargs: Any
    ) -> SimpleFMUExperimentDefinition:
        """Sets the modifiers parameters for an experiment.

        Args:
            modifiers: A dictionary of variable modifiers.

        Example::

            from modelon.impact.client import Range, Choices

            fmu = model.compile().wait()
            experiment_definition = fmu.new_experiment_definition(
                custom_function).with_modifiers({'inertia1.J': Choices(0.1, 0.9),
                'inertia2.J': Range(0.1, 0.5, 3)})

        """
        if modifiers_kwargs:
            logger.warning(
                "The argument 'modifier_kwargs' is deprecated and will be removed in a"
                " future release of the Impact Client. Please specify a "
                "dictionary of variable modifiers instead!"
            )
        modifiers = {} if modifiers is None else modifiers
        modifiers_aggregate = {**modifiers, **modifiers_kwargs}
        new = SimpleFMUExperimentDefinition(
            self._fmu,
            self._custom_function,
            self._solver_options,
            self._simulation_options,
            self._simulation_log_level,
            self._initialize_from,
        )
        new._extensions = self._extensions
        new._variable_modifiers = self._variable_modifiers
        for name, value in modifiers_aggregate.items():
            new._variable_modifiers[name] = ensure_as_modifier(value)
        return new

    def with_extensions(
        self, experiment_extensions: List[SimpleExperimentExtension]
    ) -> SimpleFMUExperimentDefinition:
        """Sets up an experiment with multiple experiment extensions.

        Args:
            experiment_extensions: "A list of experiment extension objects.
                Extension object must an instance of  SimpleExperimentExtension class.

        Example::

            fmu = model.compile().wait()
            experiment_definition = fmu.new_experiment_definition(custom_function).
            with_extensions(
                [
                    SimpleExperimentExtension().with_modifiers({'PI.k': 20}),
                    SimpleExperimentExtension().with_modifiers({'PI.k': 30}),
                ]
            )

            experiment_definition = fmu.new_experiment_definition(custom_function).
            with_extensions(
                [
                    SimpleExperimentExtension(
                        parameter_modifiers={'start_time': 0.0, 'final_time': 2.0}
                    ).with_modifiers({'PI.k': 20})
                ]
            )

        """

        assert_valid_extensions(experiment_extensions)
        exp_ext = []
        for extension in experiment_extensions:
            exp_ext.append(extension)

        new = SimpleFMUExperimentDefinition(
            self._fmu,
            self._custom_function,
            self._solver_options,
            self._simulation_options,
            self._simulation_log_level,
            self._initialize_from,
        )
        new._variable_modifiers = self._variable_modifiers
        new._extensions = self._extensions + exp_ext
        return new

    def with_cases(
        self, cases_modifiers: List[Dict[str, Any]]
    ) -> SimpleFMUExperimentDefinition:
        """Sets up an experiment with multiple cases with different variable modifiers.

        Args:
            cases_modifiers: A list of variable modifier dictionaries.
                Multiple dictionaries with variable modifiers could to added to create
                multiple cases.

        Example::

            fmu = model.compile().wait()
            experiment_definition = fmu.new_experiment_definition(
                custom_function).with_cases([{'PI.k': 20}, {'PI.k': 30}])

        """
        assert_valid_case_modifiers(cases_modifiers)

        extensions = [
            SimpleExperimentExtension().with_modifiers(modifiers)
            for modifiers in cases_modifiers
        ]
        return self.with_extensions(extensions)

    def to_dict(self) -> Dict[str, Any]:
        """Returns the experiment definition as a dictionary.

        Returns:
            A dictionary containing the experiment definition.

        Example::

            fmu = model.compile().wait()
            simulation_options = custom_function.get_simulation_options()
                .with_values(ncp=500)
            solver_options = {'atol':1e-8}
            simulate_def = fmu.new_experiment_definition(custom_function,
            solver_options, simulation_options)
            simulate_def.to_dict()

        """
        custom_function_parameters = custom_function_parameters_to_dict(
            self._custom_function.parameter_values.as_raw_dict()
        )
        variable_modifiers = {
            name: ensure_as_modifier(value)
            for name, value in self._variable_modifiers.items()
        }
        exp_dict: Dict[str, Any] = {
            "experiment": {
                "version": 3,
                "base": {
                    "model": {"fmu": {"id": self._fmu.id}},
                    "modifiers": {"variables": modifiers_to_dict(variable_modifiers)},
                    "analysis": {
                        "type": self._custom_function.name,
                        "parameters": custom_function_parameters,
                        "simulationOptions": self._simulation_options,
                        "solverOptions": self._solver_options,
                        "simulationLogLevel": self._simulation_log_level,
                    },
                },
                "extensions": [ext.to_dict() for ext in self._extensions],
            }
        }
        if isinstance(self.initialize_from, ExperimentInterface):
            exp_dict["experiment"]["base"]["modifiers"][
                "initializeFrom"
            ] = self.initialize_from.id
        elif isinstance(self.initialize_from, CaseInterface):
            exp_dict["experiment"]["base"]["modifiers"][
                "initializeFromCase"
            ] = case_to_identifier_dict(self.initialize_from)
        elif isinstance(self.initialize_from, ExternalResultInterface):
            exp_dict["experiment"]["base"]["modifiers"][
                "initializeFromExternalResult"
            ] = self.initialize_from.id
        return exp_dict

    def with_initialize_from(
        self, entity: Optional[CaseOrExperimentOrExternalResult] = None
    ) -> SimpleFMUExperimentDefinition:
        """Sets the experiment or case to initialize from for an experiment.

        Args:
            entity: An instance of Case or Experiment or ExternalResult."

        Example::

            experiment = workspace.get_experiment(experiment_id)
            fmu = model.compile().wait()
            experiment_definition = fmu.new_experiment_definition(custom_function).
            experiment_definition.with_initialize_from(experiment)

            experiment = workspace.get_experiment(experiment_id)
            case = experiment.get_case('case_1')
            fmu = model.compile().wait()
            experiment_definition = fmu.new_experiment_definition(custom_function).
            experiment_definition.with_initialize_from(case)

            result = workspace.upload_result('A.mat').wait()
            fmu = model.compile().wait()
            experiment_definition = fmu.new_experiment_definition(custom_function).
            experiment_definition.with_initialize_from(result)

        """
        validate_initialize_from(entity)
        new = SimpleFMUExperimentDefinition(
            self._fmu,
            self._custom_function,
            self._solver_options,
            self._simulation_options,
            self._simulation_log_level,
            entity,
        )
        new._variable_modifiers = self._variable_modifiers
        new._extensions = self._extensions

        return new



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from modelon.impact.client.entities.interfaces.case import CaseInterface
from modelon.impact.client.entities.interfaces.custom_function import (
    CustomFunctionInterface,
)
from modelon.impact.client.entities.interfaces.experiment import ExperimentInterface
from modelon.impact.client.entities.interfaces.external_result import (
    ExternalResultInterface,
)
from modelon.impact.client.entities.interfaces.model import ModelInterface
from modelon.impact.client.entities.interfaces.model_executable import (
    ModelExecutableInterface,
)
from modelon.impact.client.experiment_definition.interfaces.extension import (
    BaseExperimentExtension,
)
from modelon.impact.client.options import (
    CompilerOptions,
    RuntimeOptions,
    SimulationOptions,
    SolverOptions,
)

if TYPE_CHECKING:
    from modelon.impact.client.entities.case import Case
    from modelon.impact.client.entities.custom_function import CustomFunction
    from modelon.impact.client.entities.experiment import Experiment
    from modelon.impact.client.entities.external_result import ExternalResult
    from modelon.impact.client.entities.model import Model
    from modelon.impact.client.entities.model_executable import ModelExecutable
    from modelon.impact.client.experiment_definition.extension import (
        SimpleExperimentExtension,
    )

    CaseOrExperimentOrExternalResult = Union[Case, Experiment, ExternalResult]
    RuntimeOptionsOrDict = Union[RuntimeOptions, Dict[str, Any]]
    SimulationOptionsOrDict = Union[SimulationOptions, Dict[str, Any]]
    SolverOptionsOrDict = Union[SolverOptions, Dict[str, Any]]
    CompilerOptionsOrDict = Union[CompilerOptions, Dict[str, Any]]


def assert_valid_args(
    model: Optional[Model] = None,
    fmu: Optional[ModelExecutable] = None,
    custom_function: Optional[CustomFunction] = None,
    solver_options: Optional[SolverOptionsOrDict] = None,
    simulation_options: Optional[SimulationOptionsOrDict] = None,
    compiler_options: Optional[CompilerOptionsOrDict] = None,
    runtime_options: Optional[RuntimeOptionsOrDict] = None,
) -> None:
    if fmu and not isinstance(fmu, ModelExecutableInterface):
        raise TypeError("FMU must be an instance of ModelExecutable class!")
    if model and not isinstance(model, ModelInterface):
        raise TypeError("Model must be an instance of Model class!")
    if custom_function and not isinstance(custom_function, CustomFunctionInterface):
        raise TypeError("Custom_function must be an instance of CustomFunction class!")
    if solver_options is not None and not isinstance(
        solver_options, (SolverOptions, dict)
    ):
        raise TypeError(
            "Solver options must be an instance of SolverOptions class or a "
            "dictionary class object!"
        )
    if simulation_options is not None and not isinstance(
        simulation_options, (SimulationOptions, dict)
    ):
        raise TypeError(
            "Simulation options must be an instance of SimulationOptions class or"
            " a dictionary class object!"
        )
    if compiler_options is not None and not isinstance(
        compiler_options, (CompilerOptions, dict)
    ):
        raise TypeError(
            "Compiler options object must either be a dictionary or an "
            "instance of CompilerOptions class!"
        )
    if runtime_options is not None and not isinstance(
        runtime_options, (RuntimeOptions, dict)
    ):
        raise TypeError(
            "Runtime options object must either be a dictionary or an "
            "instance of RuntimeOptions class!"
        )


def validate_initialize_from(
    entity: Optional[CaseOrExperimentOrExternalResult],
) -> None:
    if entity and not isinstance(
        entity, (CaseInterface, ExperimentInterface, ExternalResultInterface)
    ):
        raise TypeError(
            "The entity argument be an instance of "
            "Case or Experiment or ExternalResult!"
        )
    if isinstance(entity, ExperimentInterface) and len(entity.get_cases()) > 1:
        raise ValueError(
            "Cannot initialize from an experiment containing multiple"
            " cases! Please specify a case object instead."
        )


def assert_valid_case_modifiers(cases_modifiers: List[Dict[str, Any]]) -> None:
    if not isinstance(cases_modifiers, list):
        raise TypeError("The case modifiers argument must be a list!")
    for case_modifier in cases_modifiers:
        if not isinstance(case_modifier, dict):
            raise TypeError(
                "The variable modifiers in the case_modifier list must be a "
                "dictionary!"
            )


def assert_valid_extensions(
    experiment_extensions: List[SimpleExperimentExtension],
) -> None:
    if not isinstance(experiment_extensions, list):
        raise TypeError("The experiment extensions argument must be a list!")
    for extension in experiment_extensions:
        if not isinstance(extension, BaseExperimentExtension):
            raise TypeError(
                "The extension object in the experiment extension list "
                "must be an instance of SimpleExperimentExtension class!"
            )



from typing import List

from modelon.impact.client import exceptions


def assert_successful_operation(
    is_successful: bool, operation_name: str = "Operation"
) -> None:
    if not is_successful:
        raise exceptions.OperationFailureError(
            f"{operation_name} failed! See the log for more info!"
        )


def assert_variable_in_result(
    variables: List[str], result_variables: List[str]
) -> None:
    add = set(variables) - set(result_variables)
    if add:
        raise ValueError(
            f"Variable(s) '{', '.join(add)}' {'are' if len(add)>1 else 'is'} not"
            " present in the result"
        )



from abc import ABC, abstractmethod
from typing import Any, Dict, Optional


class ExpansionAlgorithm(ABC):
    """Base class for an expansion algorithm."""

    @abstractmethod
    def __str__(self) -> str:
        "Returns a string representation of the expansion algorithm"

    @abstractmethod
    def get_parameters_as_dict(self) -> Optional[Dict[str, Any]]:
        "Returns parameters as a dictionary"
